/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsource.h"
#include "gdammp3.h"
#include "gdamexternal.h"
#include <gdam/gdambuffer.h>
#include <gdam/gdamarg.h>
#include <gdam/gdam-protocol.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>

static void gdam_mp3_start(GdamMp3* mp3, GdamTime start, GdamTime length);

static void gdam_mp3_set_filename(GtkObject* object, GdamArg* arg)
{
	GdamMp3* mp3 = GDAM_MP3(object);
	mp3->filename = g_strdup(GDAM_ARG_STRING(arg));
}
static void gdam_mp3_get_filename(GtkObject* object, GdamArg* arg)
{
	GdamMp3* mp3 = GDAM_MP3(object);
	GDAM_ARG_STRING(arg) = g_strdup(mp3->filename);
}

static void gdam_mp3_set_position(GtkObject* object, GdamArg* arg)
{
	GdamMp3* mp3 = GDAM_MP3(object);
	mp3->start_time = GDAM_ARG_INT64(arg);
}
static void gdam_mp3_get_position(GtkObject* object, GdamArg* arg)
{
	GdamExternal* external = GDAM_EXTERNAL(object);
	GdamMp3* mp3 = GDAM_MP3(object);
	if (external->reader_source == 0)
		GDAM_ARG_INT64(arg) = mp3->start_time;
	else
		GDAM_ARG_INT64(arg) = external->samples_output;
}

static void gdam_mp3_set_downsample(GtkObject* object, GdamArg* arg)
{
	GdamMp3* mp3 = GDAM_MP3(object);
	int downsample = GDAM_ARG_INT(arg);
	if (downsample == 1 || downsample == 2 || downsample == 4) {
		mp3->downsample = downsample;
	} else {
		gdam_debug("invalid downsample %d, must be 1,2,or 4",
		           downsample);
	}
}
static void gdam_mp3_get_downsample(GtkObject* object, GdamArg* arg)
{
	GdamMp3* mp3 = GDAM_MP3(object);
	GDAM_ARG_INT(arg) = mp3->downsample;
}


static void gdam_mp3_done_constructing(GdamSource* source)
{
	GdamMp3* mp3 = GDAM_MP3(source);
	gdam_mp3_start(mp3,
	               mp3->start_time,
		       GDAM_EXTERNAL(mp3)->samples_remaining);
}

static void gdam_mp3_class_init(GtkObjectClass* object_class)
{
	GDAM_SOURCE_CLASS(object_class)->done_constructing
				= gdam_mp3_done_constructing;
	gdam_arg_register(object_class,
	                  "filename",
			  GDAM_ARG_TYPE_STRING,
			  gdam_mp3_set_filename,
			  gdam_mp3_get_filename,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "position",
			  GDAM_ARG_TYPE_INT64,
			  gdam_mp3_set_position,
			  gdam_mp3_get_position,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "downsample",
			  GDAM_ARG_TYPE_INT,
			  gdam_mp3_set_downsample,
			  gdam_mp3_get_downsample,
			  GDAM_ARG_VITAL);
}

static void gdam_mp3_init(GdamMp3* mp3)
{
	mp3->downsample = 1;
}

static GtkObjectClass*  parent_class = NULL;

GtkType gdam_mp3_get_type()
{
	static GtkType mp3_type = 0;
	if (!mp3_type) {
		static const GtkTypeInfo mp3_info = {
			"GdamMp3",
			sizeof(GdamMp3),
			sizeof(GdamMp3Class),
			(GtkClassInitFunc) gdam_mp3_class_init,
			(GtkObjectInitFunc) gdam_mp3_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_EXTERNAL);
		mp3_type = gtk_type_unique(GDAM_TYPE_EXTERNAL, &mp3_info);
	}
	return mp3_type;
}

#define SAMPLES_PER_FRAME 1152


static void
gdam_mp3_start(GdamMp3* mp3, GdamTime start, GdamTime length)
{
	int start_frame;
	int miss;
	char** command;
	int i = 0;
	GdamExternal* external = GDAM_EXTERNAL(mp3);
	if (start < 0) {
		external->zero_time = (int) (-start);
		start = 0;
	}
	start_frame = (int)(start / SAMPLES_PER_FRAME);
	command = g_new(char*, 32);
	command[i++] = g_strdup("mpg123");
	command[i++] = g_strdup("-q");
	command[i++] = g_strdup("-r");
	command[i++] = g_strdup("44100");
	command[i++] = g_strdup("--stereo");
	command[i++] = g_strdup("-s");
	command[i++] = g_strdup("-k");
	command[i++] = g_strdup_printf("%d", start_frame);
	command[i++] = g_strdup(mp3->filename);
	command[i++] = NULL;
	if (length != -1) {
		command[i++] = g_strdup("-n");
		command[i++] = g_strdup_printf("%d",
				    (int)(length / SAMPLES_PER_FRAME) + 2);
	}
	if (mp3->downsample == 2) {
		command[i++] = g_strdup("-2");
	} else if (mp3->downsample == 4) {
		command[i++] = g_strdup("-4");
	}
	command[i++] = g_strdup(mp3->filename);
	command[i] = NULL;
	gdam_external_run_command(external, command);
	miss = (int)(start - (GdamTime)start_frame*SAMPLES_PER_FRAME);
	g_assert(miss >= 0);
	if (miss > 0)
		gdam_source_miss_data(GDAM_SOURCE(mp3), miss);

	external->samples_output = start;

	/* Since samples_remaining happens to be the number used
	 * for `unlimited' in both cases, ... */
	external->samples_remaining = length;
	g_strfreev(command);
}

GdamMp3* gdam_mp3_new(const char* filename, GdamTime start, GdamTime len)
{
	GdamMp3* mp3 = GDAM_MP3(gtk_type_new(GDAM_TYPE_MP3));
	if (start < 0) {
		mp3->external.zero_time = -start;
		start = 0;
	}
	mp3->filename = g_strdup(filename);
	gdam_mp3_start(mp3, start, len);
	return mp3;
}
